import tensorflow as tf
from tensorflow.keras.layers import LSTM, Dense, Embedding


class Encoder(tf.keras.Model):
  def __init__(self,
               vocab_size,
               embedding_dim,
               enc_units,
               n_layers):
    super(Encoder, self).__init__()
    self.n_layers = n_layers # used for stacked LSTM, n_layers=1: single layer
    self.enc_units = enc_units # number of hidden units
    self.embedding = Embedding(vocab_size, embedding_dim)

    # suggestion for being used for attention
    self.W1 = Dense(self.enc_units)
    self.W2 = Dense(self.enc_units)

    # TODO: define RNN(s) below
    # for stacked and multi layers
    # suggestion: to make it cover for the simple case as well.
    self.forw = None # forward RNN
    self.back = None  # backward RNN (only used for Bidirectional RNN)

  def call(self, x):
    x = self.embedding(x)
    f, b = x, x

    # TODO: inference for RNN
    # suggestion to make it work for multilayers and bidrectional as well
    # forward, and backward and combine them

    # TODO: attention
    # below is for Luong (concat type) attention
    keys = None
    values = None

    return keys, values


class Decoder(tf.keras.Model):
  def __init__(self,
               vocab_size,
               embedding_dim,
               dec_units,
               n_layers):
    super(Decoder, self).__init__()
    self.dec_units = dec_units # number of hidden units
    self.n_layers = n_layers  # used for stacked LSTM, n_layers=1: single layer
    self.embedding = Embedding(vocab_size, embedding_dim)

    # TODO: define a unidirectional RNN here
    self.lstms = None

    # TODO: define a respective weight for (suggested: Luong) attention
    self.V = None

    self.fc = Dense(vocab_size)

  def call(self, x, hidden, keys, values):
    # keys/values shape == (batch_size, max_length, key_size)
    x = self.embedding(x)

    # TODO: inference, collect output and states
    states = []
    for i in range(self.n_layers):
      pass

    # ------------------------------------------------------------------
    #        ATTENTION
    # implement attention of your own, no need to follow our suggestion
    # ------------------------------------------------------------------

    # TODO: calculate attention scores and normalize it with softmax
    # [B, 1, dec_units]
    # query vector = x
    scores_normalized = None

    # TODO: Calculate the weighted average of the attention inputs
    # according to the scores to refine decoder's outputs with attention
    context = None
    output = None
    # ------------------------------------------------------------------

    # expand to vocab size to decode
    output = self.fc(output)

    return output, states


def model(data_dict,
          checkpoint_dir='./checkpiont',
          emb_dim=256,
          n_hidden=1024,
          n_dec_layers=1,
          n_enc_layers=1,
          max_ckpt_to_keep=3,
          lr=1e-3,
          ):
  """WARNING: You don't need to change anything on this method. 
  """
  encoder = Encoder(data_dict['src_vocab_size'],
                    emb_dim,
                    n_hidden,
                    n_enc_layers)
  decoder = Decoder(data_dict['tgt_vocab_size'],
                    emb_dim,
                    n_hidden,
                    n_dec_layers)
  optimizer = tf.keras.optimizers.Adam(lr)

  # checkpoint will not only save the model but also the optimizer
  checkpoint = tf.train.Checkpoint(optimizer=optimizer,
                                   encoder=encoder,
                                   decoder=decoder)
  manager = tf.train.CheckpointManager(checkpoint,
                                       checkpoint_dir,
                                       max_to_keep=max_ckpt_to_keep)
  return {
    'encoder': encoder,
    'decoder': decoder,
    'train_ckpt': checkpoint,
    'ckpt_manager': manager,
    'optim': optimizer,
  }
